"""toil_container containers tests."""

from os.path import join
from os.path import abspath
from os.path import dirname
import os
import argparse
import subprocess

from docker.errors import ImageNotFound
import docker
import pytest

from toil.common import Toil
from toil.job import Job

from toil_container import __version__
from toil_container import utils
from toil_container.containers import Container
from toil_container.jobs import ContainerCallJob


ROOT = abspath(join(dirname(__file__), ".."))


def build_docker_image(image_tag):
    """Build image from Dockerfile if not created yet."""
    try:
        client = docker.from_env()
        client.images.get(image_tag)
    except ImageNotFound:
        client.images.build(path=ROOT, rm=True, tag=image_tag)


@pytest.mark.skipif(
    not utils.is_docker_available(),
    reason="docker is not available."
    )
def test_docker_container():
    """
    Test the docker image is created properly and it executes the entrypoint
    as expected.

        docker build -t test-image .
        docker run -it test-image --version
    """
    image_tag = "test-docker"
    cmd = [
        "python",
        "-c",
        'from toil_container import __version__; print(__version__)'
        ]
    build_docker_image(image_tag)

    container = Container().docker_call(
        image_tag,
        cmd=cmd,
        check_output=True,
        )

    assert __version__ in container.decode()


@pytest.mark.skipif(
    not utils.is_singularity_available(),
    reason="singularity is not available."
    )
def test_singularity_container():
    """
    Test the singularity image exists and its executable.

        singularity exec <test-image.simg> <command-parameters>
    """
    singularity_image = "docker://ubuntu"
    cmd = ["cat", "/etc/os-release"]

    # Create call
    output = Container().singularity_call(
        singularity_image,
        cmd=cmd,
        check_output=True
        )

    assert "VERSION" in output.decode()



# Toil Jobs and Options for testing
class ContainerizedCheckCallJob(ContainerCallJob):
    """
    Job created to test that check_call is used correctly by docker
    and singularity.
    """
    cmd = ["pwd"]
    cwd = None
    env = {}

    def run(self, jobStore):
        """Saves a Hello message in a file."""
        return self.check_call(self.cmd, cwd=self.cwd, env=self.env)


class ContainerizedCheckOutputJob(ContainerCallJob):
    """
    Job created to test that check_output is used correctly by docker
    and singularity.
    """
    cmd = ["pwd"]
    cwd = None
    env = {}

    def run(self, jobStore):
        """Saves a Hello message in a file."""
        return self.check_output(self.cmd, cwd=self.cwd, env=self.env)


def get_toil_test_parser():
    """Get test pipeline configuration using argparse."""
    # Add Toil options.
    parser = argparse.ArgumentParser()
    Job.Runner.addToilOptions(parser)

    # Parameters to run with docker or singularity
    settings = parser.add_argument_group("To run with docker or singularity:")

    settings.add_argument("--docker", default=False)
    settings.add_argument("--singularity", required=False)
    settings.add_argument("--shared-fs", required=False)

    return parser


def set_container_arguments(args, container_tool=None):
    """Add toil arguments necessary to run in containers."""
    shared_fs = os.environ["TEST_SHARED_FS"]

    if container_tool == "docker":
        image_tag = "test-toil"
        build_docker_image(image_tag)

        args += [
            "--docker", image_tag,
            "--shared-fs", shared_fs,
            ]

    elif container_tool == "singularity":
        singularity_image = "docker://ubuntu"

        args += [
            "--singularity", singularity_image,
            "--shared-fs", shared_fs,
            ]

    return args


def run_job_in_control_env(tmpdir, container_tool=None):
    """
    Run a Toil job in container with the option --singularity or --docker
    and --shared-fs SHARED-DIRECTORY to test the singularity wrapper
    is executing correctly the command inside the container.
    """
    # Create options for job
    workdir = join(str(tmpdir))
    jobstore = join(str(tmpdir), "jobstore")

    # Define arguments
    args = [jobstore, "--workDir", workdir]
    args = set_container_arguments(args, container_tool)
    parser = get_toil_test_parser()
    options = parser.parse_args(args)

    # Create jobs
    job_call = ContainerizedCheckCallJob(
        options=options,
        unitName="Check call pwd",
        )

    job_output = ContainerizedCheckOutputJob(
        options=options,
        unitName="Check output pwd",
        )

    # Make sure that cwd functionality in check_output works
    job_output.cwd = "/home"
    std_output = job_output.run(jobstore)
    assert "/home" in std_output

    # Make sure that cwd functionality in check_call works
    std_call = job_call.run(jobstore)
    assert 0 == std_call

    # Make sure workDir is used as the tmp directory inside the container
    # and that an ENV variable is passed to the container system call.
    message = "hello World"
    job_output.env = {"ISLAND": message}

    out_file = "bottle.txt"
    tmp_file_in_container = join(os.sep, "tmp", out_file)

    if container_tool == "docker":
        tmp_file_in_workdir = join(workdir, out_file)
    elif container_tool == "singularity":
        tmp_file_in_workdir = join(workdir, "tmp", out_file)
    else:
        tmp_file_in_workdir = join(workdir, out_file)
        tmp_file_in_container = tmp_file_in_workdir

    job_output.cmd = [
        "/bin/bash",
        "-c",
        "echo $ISLAND > {}".format(tmp_file_in_container)
        ]
    job_output.run(jobstore)

    with open(tmp_file_in_workdir) as f:
        assert message in f.read()


def run_parallel_jobs(tmpdir, container_tool=None):
    """
    Make sure parallel jobs work.
                    |
                parent_job
              /            \
    child_job_1             child_job_2
              \            /
                 tail_job
    """
    # Create options for job
    workdir = join(str(tmpdir))
    jobstore = join(str(tmpdir), "jobstore")

    args = [jobstore, "--workDir", workdir]
    args = set_container_arguments(args, container_tool)

    parser = get_toil_test_parser()
    options = parser.parse_args(args)

    # Create jobs
    parent_job = ContainerizedCheckOutputJob(
        options=options,
        unitName="Parent job",
        )
    child_job_1 = ContainerizedCheckOutputJob(
        options=options,
        unitName="Parallel Job 1",
        )
    child_job_2 = ContainerizedCheckOutputJob(
        options=options,
        unitName="Parallel Job 2",
        )
    tail_job = ContainerizedCheckOutputJob(
        options=options,
        unitName="Last Followon Job",
        )

    # Assign job commands
    out_file = "bottle.txt"
    tmp_file_in_container = join(os.sep, "tmp", out_file)

    if container_tool == "docker":
        tmp_file_in_workdir = join(workdir, out_file)
    elif container_tool == "singularity":
        tmp_file_in_workdir = join(workdir, "tmp", out_file)
    else:
        tmp_file_in_workdir = join(workdir, out_file)
        tmp_file_in_container = tmp_file_in_workdir

    base_cmd = ["/bin/bash", "-c"]

    parent_job.cmd = base_cmd + [
        "echo job1 >> {}".format(tmp_file_in_container)
        ]
    child_job_1.cmd = base_cmd + [
        "echo job2 >> {}".format(tmp_file_in_container)
        ]
    child_job_2.cmd = base_cmd + [
        "echo job3 >> {}".format(tmp_file_in_container)
        ]
    tail_job.cmd = base_cmd + [
        "echo job4 >> {}".format(tmp_file_in_container)
        ]

    # Create DAG
    parent_job.addChild(child_job_1)
    parent_job.addChild(child_job_2)
    parent_job.addFollowOn(tail_job)

    # Run jobs
    with Toil(options) as pipe:
        pipe.start(parent_job)

    # Test the output
    with open(tmp_file_in_workdir) as f:
        assert len(f.readlines()) == 4


def test_subprocess_toil_single_jobs(tmpdir):
    """Test to check subprocess calls work."""
    run_job_in_control_env(tmpdir)


@pytest.mark.skipif(
    not utils.is_docker_available(),
    reason="docker is not available."
    )
def test_docker_toil_single_jobs(tmpdir):
    """
    Test to check docker is setting correctly the /tmp dir,
    the working dir and the ENV variables inside the container.
    """
    run_job_in_control_env(tmpdir, container_tool="docker")


@pytest.mark.skipif(
    not utils.is_singularity_available(),
    reason="singularity is not available."
    )
def test_singularity_toil_single_jobs(tmpdir):
    """
    Test to check singularity is setting correctly the /tmp dir,
    the working dir and the ENV variables inside the container.
    """
    run_job_in_control_env(tmpdir, container_tool="singularity")


def test_subprocess_toil_parallel_jobs(tmpdir):
    """ Test to check parallel subprocess calls work."""
    run_parallel_jobs(tmpdir)


@pytest.mark.skipif(
    not utils.is_docker_available(),
    reason="docker is not available."
    )
def test_docker_toil_parallel_jobs(tmpdir):
    """ Test to check parallel docker containers run."""
    run_parallel_jobs(tmpdir, container_tool="docker")


@pytest.mark.skipif(
    not utils.is_singularity_available(),
    reason="singularity is not available."
    )
def test_singularity_toil_parallel_jobs(tmpdir):
    """ Test to check parallel singularity containers run."""
    run_parallel_jobs(tmpdir, container_tool="singularity")




"""toil_container tests utils."""

from cStringIO import StringIO
import sys


class Capturing(list):

    """
    Capture stdout of a function call.

    See: https://stackoverflow.com/questions/16571150

    Example:

        with Capturing() as output:
            do_something(my_object)
    """

    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout



"""toil_container specific exceptions."""


class ToilContainerException(Exception):

    """A base exception for toil_container."""


class DockerNotAvailableError(ToilContainerException):

    """A class to raise when docker is not available."""


class SingularityNotAvailableError(ToilContainerException):

    """A class to raise when singularity is not available."""



"""toil_container parsers tests."""

import click
import pytest

from toil_container import parsers

from .utils import Capturing


def check_help_toil(parser):
    with Capturing() as without_toil:
        try:
            parser.parse_args(["--help"])
        except SystemExit:
            pass

    with Capturing() as with_toil:
        try:
            parser.parse_args(["--help-toil"])
        except SystemExit:
            pass

    with_toil = "\n".join(with_toil)
    without_toil = "\n".join(without_toil)

    # By default toil options shouldn't be printed.
    assert "toil core options" not in without_toil
    assert "TOIL OPTIONAL ARGS" in without_toil
    assert "toil arguments" in without_toil

    # Check that toil options were added by default.
    assert "toil core options" not in without_toil
    assert "toil core options" in with_toil


def check_container_options(parser, tmpdir):
    shared_fs = tmpdir.mkdir("shared_fs")
    work_dir = tmpdir.mkdir("workdir")
    image = tmpdir.join("test.img")
    jobstore = tmpdir.join("jobstore")
    image.write("not empty")
    parser = parsers.ContainerArgumentParser()

    # Can't pass docker and singularity at the same time.
    args = [
        "--singularity", image.strpath,
        "--docker", image.strpath,
        jobstore.strpath,
    ]

    # By default container options should be added in both helps.
    assert "container arguments" in parser.format_help()
    assert "container arguments" in parser.format_help()

    with pytest.raises(click.UsageError) as error:
        parser.parse_args(args)

    assert "You can't pass both" in str(error.value)

    # shared-fs only used for containers.
    args = [
        "--shared-fs", shared_fs.strpath,
        "--workDir", work_dir.strpath,
        jobstore.strpath,
    ]

    with pytest.raises(click.UsageError) as error:
        parser.parse_args(args)

    assert "--shared-fs should be used only " in str(error.value)

    # workDir must be inside shared-fs.
    args = [
        "--singularity", image.strpath,
        "--shared-fs", shared_fs.strpath,
        "--workDir", work_dir.strpath,
        jobstore.strpath,
    ]

    with pytest.raises(click.UsageError) as error:
        parser.parse_args(args)

    assert "The --workDir must be available " in str(error.value)


def test_help_toil():
    parser = parsers.ToilShortArgumentParser()
    check_help_toil(parser)


def test_help_toil_container():
    parser = parsers.ContainerShortArgumentParser()
    check_help_toil(parser)


def test_container_parse_args(tmpdir):
    parser = parsers.ContainerArgumentParser()
    check_container_options(parser, tmpdir)


def test_container_parse_args_short(tmpdir):
    parser = parsers.ContainerShortArgumentParser()
    check_container_options(parser, tmpdir)



"""
Library to manage docker and singularity calls through the containers tools.

Docker uses a API for python, while singularity API doesn't support python2.

Based on the docker implementation of:
https://github.com/BD2KGenomics/toil/blob/master/src/toil/lib/docker.py

Based on the singularity implementation of:
https://github.com/vgteam/toil--bindg/blob/master/src/toil_vg/singularity.py

Docker images can be run by prefacing the input image with docker://
In this case, Singularity will download, convert, and cache the image on the
fly. This cache can be set with SINGULARITY_CACHEDIR, and defaults to the
user's home directory. This cache can be a major bottleneck when repeatedly
more different images than it can hold (not very many). So for this type of
usage pattern (concurrent or short consecutive calls to different images),
it is best to run Singularity images natively.
"""

import logging
import uuid
import subprocess

from docker.errors import ContainerError
from docker.errors import ImageNotFound
import docker

from toil_container import utils

_LOGGER = logging.getLogger(__name__)


class Container(object):

    """Class to manage docker and singularity calls."""

    def __init__(self):
        """Assign attributes unique to each container."""
        if utils.is_docker_available():
            self.docker_client = docker.from_env()
            self.container_name = ""

    def docker_call(
            self,
            image,
            cmd=None,
            cwd=None,
            env=None,
            check_output=None,
            working_dir=None,
            shared_fs=None):
        """
        Execute parameters in a docker container via docker-python API.

        See: https://docker-py.readthedocs.io/en/stable/

        Arguments:
            image (str): name of the singularity image.
            cmd (list): list of command line arguments passed to the tool.
            cwd (str): current working directory.
            env (dict): environment variables to set inside container.
            check_output (bool): check_output or check_call behavior.
            working_dir (dict): directory where commands will be run.
            shared_fs (str): file shared dir to bind volumens inside container.

        Returns:
            str: (check_output=True) stdout of the system call.
            int: (check_output=False) 0 if call succeed else raise error.

        Raises:
            CalledProcessorError: if the container invocation returns a
            non-zero exit code.
        """
        utils.is_docker_available(raise_error=True)

        if not self.container_name:
            self.container_name = self._get_container_name(image)

        docker_parameters = {}
        docker_parameters["command"] = cmd
        docker_parameters["detach"] = check_output
        docker_parameters["entrypoint"] = ""
        docker_parameters["environment"] = env or {}
        docker_parameters["name"] = self.container_name
        docker_parameters["volumes"] = {}

        # Set parameters for managing directories if options are defined
        if shared_fs:
            docker_parameters["volumes"][shared_fs] = {
                "bind": shared_fs,
                "mode": "rw"
                }

        if working_dir:
            docker_parameters["volumes"][working_dir] = {
                "bind": "/tmp",
                "mode": "rw"
                }

        if cwd:
            docker_parameters["working_dir"] = cwd

        exceptions = (
            ContainerError,
            ImageNotFound,
            subprocess.CalledProcessError
            )

        try:
            container = self.docker_client.containers.run(
                image,
                **docker_parameters
                )

            # If detached wait to get the system exit to get logs before remove.
            if check_output:
                container.wait()
                output = container.logs()
            else:
                output = 0
            self._prune_docker_container(self.container_name)

            return output

        except exceptions as stderr:
            self._prune_docker_container(self.container_name)
            raise subprocess.CalledProcessError(0, cmd=cmd, output=stderr)

    @staticmethod
    def singularity_call(
            image,
            cmd=None,
            cwd=None,
            env=None,
            check_output=None,
            working_dir=None,
            shared_fs=None):
        """
        Execute parameters in a singularity container via subprocess.

        The resulting command is in the format:

            singularity -q exec
                <list-of-singularity-params>
                <singularity-image>
                <list-of-cmd-params-executed-inside-container>

        Arguments:
            image (str): name of the singularity image.
            cmd (list): list of command line arguments passed to the tool.
            cwd (str): current working directory.
            env (dict): environment variables to set inside container.
            check_output (bool): check_output or check_call behavior.
            working_dir (dict): directory where commands will be run.
            shared_fs (str): file shared dir to bind volumens inside container.

        Returns:
            str: (check_output=True) stdout of the system call.
            int: (check_output=False) 0 if call succeed else non-0.

        Raises:
            CalledProcessorError: if the container invocation returns a
            non-zero exit code.
        """
        singularity_command = utils.is_singularity_available(
            raise_error=True,
            path=True
            )
        singularity_parameters = []

        # Set parameters for managing directories if options are defined
        if shared_fs:
            singularity_parameters += [
                "--bind", "{fs}:{fs}".format(fs=shared_fs)
                ]
        if working_dir:
            singularity_parameters += [
                "--contain",
                "--workdir", working_dir
                ]
        if cwd:
            singularity_parameters += ["--pwd", cwd]

        # Setup the outgoing subprocess call for singularity
        command = [singularity_command, "-q", "exec"]
        command += singularity_parameters or []
        command += [image]
        command += cmd or []

        subprocess_kwargs = {}

        # Singularity inherits the subprocess environment
        subprocess_kwargs["env"] = env

        if check_output:
            call_method = subprocess.check_output
        else:
            call_method = subprocess.check_call

        _LOGGER.info("Calling singularity with %s", repr(command))
        out = call_method(command, **subprocess_kwargs)
        return out

    @staticmethod
    def _get_container_name(image):
        """
        Create a unique name for the container.

        Arguments:
            image (str): name of the image used to run the container.

        Returns:
            str: a unique string of image name plus a unique random identifier.
        """
        return image.replace(":latest", "") + "-" + str(uuid.uuid4())

    def _prune_docker_container(self, container_name):
        """
        Remove running container.

        Arguments:
            container_name (str): name of unique container that is running.
        """
        if not container_name:
            container_name = self.container_name
        container = self.docker_client.containers.get(container_name)
        container.stop()
        container.remove()



"""toil_container setup.py."""

from os.path import abspath
from os.path import dirname
from os.path import join
import io
import json

from setuptools import find_packages
from setuptools import setup


def read(path, **kwargs):
    """Return content of a file."""
    return io.open(path, encoding=kwargs.get("encoding", "utf8")).read()

# make sure we use absolute paths
ROOT = abspath(dirname(__file__))

# please put setup keywords in the setup.json to keep this file clean
with open(join(ROOT, "setup.json"), "r") as f:
    SETUP = json.load(f)

setup(
    # load description from README
    long_description=read(join(ROOT, "README.md")),

    # in combination with MANIFEST.in, non-python files included
    # inside the toil_container will be copied to the
    # site-packages and wheels installation directory
    include_package_data=True,

    # return a list all Python packages found within the ROOT directory
    packages=find_packages(),

    # The version is only defined in one place
    version=read(join(ROOT, "toil_container", "VERSION")).strip(),

    # pass parameters loaded from setup.json including author and version
    **SETUP
)



"""toil_container jobs."""

import subprocess

from toil.job import Job

from toil_container.containers import Container


class ContainerCallJob(Job):

    """
    A job class with abstract methods for system calls.

    This class includes abstract methods `check_call` and `check_output` which
    will use `toil` or `singularity` containers to execute system calls.

    In order to do this, the `options` namespace must have the properties
    `docker_image_path` or `singularity_image_path`, else python's `subprocess`
    will be used for system calls.

    Use `toil_container.ContainerOptionsParser` to include `--docker` and
    `--singularity` in the options namespace!
    """

    def __init__(self, options, *args, **kwargs):
        """
        Set `options` namespace as an attribute.

        Arguments:
            options (object): an `argparse.Namespace` object.
            args (list): positional arguments to be passed to `toil.job.Job`.
            kwargs (dict): key word arguments to be passed to `toil.job.Job`.
        """
        self.options = options
        super(ContainerCallJob, self).__init__(*args, **kwargs)

    def check_call(self, cmd, cwd=None, env=None):
        """
        Wrap the subprocess.check_call, if any container tool was chosen.

        Arguments:
            cmd (list): list of command line arguments passed to the tool.
            cwd (str): current working directory.
            env (dict): environment variables to set inside container.

        Returns:
            int: 0 if call succeed else raise error.
        """
        if getattr(self.options, "singularity", None):
            return Container().singularity_call(
                self.options.singularity,
                cmd=cmd,
                cwd=cwd,
                env=env,
                check_output=False,
                working_dir=self.options.workDir,
                shared_fs=self.options.shared_fs,
                )
        elif getattr(self.options, "docker", None):
            return Container().docker_call(
                self.options.docker,
                cmd=cmd,
                cwd=cwd,
                env=env,
                check_output=False,
                working_dir=self.options.workDir,
                shared_fs=self.options.shared_fs,
                )
        return subprocess.check_call(
            cmd,
            cwd=cwd,
            env=env
            )

    def check_output(self, cmd, cwd=None, env=None):
        """
        Wrap the subprocess.check_output, if any container tool was chosen.

        Arguments:
            cmd (list): list of command line arguments passed to the tool.
            cwd (str): current working directory.
            env (dict): environment variables to set inside container.

        Returns:
            str: stdout of the system call.
        """
        if self.options.singularity:
            return Container().singularity_call(
                self.options.singularity,
                cmd=cmd,
                cwd=cwd,
                env=env,
                check_output=True,
                working_dir=self.options.workDir,
                shared_fs=self.options.shared_fs,
                )
        elif self.options.docker:
            return Container().docker_call(
                self.options.docker,
                cmd=cmd,
                cwd=cwd,
                env=env,
                check_output=True,
                working_dir=self.options.workDir,
                shared_fs=self.options.shared_fs,
                )
        return subprocess.check_output(
            cmd,
            cwd=cwd,
            env=env
            )
